"""Whisper CLI - Advanced voice-to-text transcription tool."""

__version__ = "0.0.1"  # This will be set by the release workflow
__author__ = "Patrick"
__description__ = "Enhanced Whisper CLI with user experience features"
